from collections.abc import Sequence
from typing import Any, overload

from cognite.client._api.time_series import SortSpec, TimeSeriesAPI
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.data_classes.filters import Filter
from cognite.client.data_classes.time_series import TimeSeriesFilter, TimeSeriesSort
from cognite.client.utils._auxiliary import split_into_chunks, unpack_items_in_payload
from cognite.client.utils._concurrency import execute_tasks
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils._validation import prepare_filter_sort, process_asset_subtree_ids, process_data_set_ids
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.extended_timeseries import ExtendedTimeSeries, ExtendedTimeSeriesList
from cognite_toolkit._cdf_tk.client.data_classes.pending_instances_ids import PendingInstanceId


class ExtendedTimeSeriesAPI(TimeSeriesAPI):
    """Extended TimeSeriesAPI to include pending ID methods."""

    @overload
    def set_pending_ids(
        self, instance_id: NodeId | tuple[str, str], id: int | None = None, external_id: str | None = None
    ) -> ExtendedTimeSeries: ...

    @overload
    def set_pending_ids(self, instance_id: Sequence[PendingInstanceId]) -> ExtendedTimeSeriesList: ...

    def set_pending_ids(
        self,
        instance_id: NodeId | tuple[str, str] | Sequence[PendingInstanceId],
        id: int | None = None,
        external_id: str | None = None,
    ) -> ExtendedTimeSeries | ExtendedTimeSeriesList:
        """Set a pending identifier for a time series.

        Args:
            instance_id: The pending instance ID to set.
            id: The ID of the time series.
            external_id: The external ID of the time series.

        Returns:
            ExtendedTimeSeries: If a single instance ID is provided, returns the updated ExtendedTimeSeries object.
        """
        if isinstance(instance_id, NodeId) or (
            isinstance(instance_id, tuple)
            and len(instance_id) == 2
            and isinstance(instance_id[0], str)
            and isinstance(instance_id[1], str)
        ):
            return self._set_pending_ids([PendingInstanceId(instance_id, id=id, external_id=external_id)])[0]  # type: ignore[return-value, arg-type]
        elif isinstance(instance_id, Sequence) and all(isinstance(item, PendingInstanceId) for item in instance_id):
            return self._set_pending_ids(instance_id)  # type: ignore[arg-type]
        else:
            raise TypeError(
                "instance_id must be a NodeId, a tuple of (str, str), or a sequence of PendingIdentifier objects."
            )

    def _set_pending_ids(self, identifiers: Sequence[PendingInstanceId]) -> ExtendedTimeSeriesList:
        """Set multiple pending identifiers for time series.

        Args:
            identifiers: A sequence of PendingIdentifier objects containing the pending instance IDs and optional IDs or external IDs.

        Returns:
            ExtendedTimeSeriesList: A list of ExtendedTimeSeries objects with the updated pending identifiers.
        """
        tasks = [
            {
                "url_path": f"{self._RESOURCE_PATH}/set-pending-instance-ids",
                "json": {
                    "items": [identifier.dump(camel_case=True) for identifier in id_chunk],
                },
                "api_subversion": "alpha",
            }
            for id_chunk in split_into_chunks(list(identifiers), 1000)
        ]
        tasks_summary = execute_tasks(
            self._post,
            tasks,
            max_workers=self._config.max_workers,
            fail_fast=True,
        )
        tasks_summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=unpack_items_in_payload,
        )

        retrieved_items = tasks_summary.joined_results(lambda res: res.json()["items"])

        return ExtendedTimeSeriesList._load(retrieved_items, cognite_client=self._cognite_client)

    def retrieve(
        self, id: int | None = None, external_id: str | None = None, instance_id: NodeId | None = None
    ) -> ExtendedTimeSeries | None:
        """`Retrieve a single time series by id. <https://developer.cognite.com/api#tag/Time-series/operation/getTimeSeriesByIds>`_

        Args:
            id (int | None): ID
            external_id (str | None): External ID
            instance_id (NodeId | None): Instance ID

        Returns:
            TimeSeries | None: Requested time series or None if it does not exist.

        Examples:

            Get time series by id:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.time_series.retrieve(id=1)

            Get time series by external id:

                >>> res = client.time_series.retrieve(external_id="1")
        """
        identifiers = IdentifierSequence.load(ids=id, external_ids=external_id, instance_ids=instance_id).as_singleton()
        return self._retrieve_multiple(
            list_cls=ExtendedTimeSeriesList,
            resource_cls=ExtendedTimeSeries,
            identifiers=identifiers,
            api_subversion="alpha",
        )

    def retrieve_multiple(
        self,
        ids: Sequence[int] | None = None,
        external_ids: SequenceNotStr[str] | None = None,
        instance_ids: Sequence[NodeId] | None = None,
        ignore_unknown_ids: bool = False,
    ) -> ExtendedTimeSeriesList:
        """`Retrieve multiple time series by id. <https://developer.cognite.com/api#tag/Time-series/operation/getTimeSeriesByIds>`_

        Args:
            ids (Sequence[int] | None): IDs
            external_ids (SequenceNotStr[str] | None): External IDs
            instance_ids (Sequence[NodeId] | None): Instance IDs
            ignore_unknown_ids (bool): Ignore IDs and external IDs that are not found rather than throw an exception.

        Returns:
            TimeSeriesList: The requested time series.

        Examples:

            Get time series by id:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.time_series.retrieve_multiple(ids=[1, 2, 3])

            Get time series by external id:

                >>> res = client.time_series.retrieve_multiple(external_ids=["abc", "def"])
        """
        identifiers = IdentifierSequence.load(ids=ids, external_ids=external_ids, instance_ids=instance_ids)
        return self._retrieve_multiple(
            list_cls=ExtendedTimeSeriesList,
            resource_cls=ExtendedTimeSeries,
            identifiers=identifiers,
            ignore_unknown_ids=ignore_unknown_ids,
            api_subversion="alpha",
        )

    def list(
        self,
        name: str | None = None,
        unit: str | None = None,
        unit_external_id: str | None = None,
        unit_quantity: str | None = None,
        is_string: bool | None = None,
        is_step: bool | None = None,
        asset_ids: Sequence[int] | None = None,
        asset_external_ids: SequenceNotStr[str] | None = None,
        asset_subtree_ids: int | Sequence[int] | None = None,
        asset_subtree_external_ids: str | SequenceNotStr[str] | None = None,
        data_set_ids: int | Sequence[int] | None = None,
        data_set_external_ids: str | SequenceNotStr[str] | None = None,
        metadata: dict[str, Any] | None = None,
        external_id_prefix: str | None = None,
        created_time: dict[str, Any] | None = None,
        last_updated_time: dict[str, Any] | None = None,
        partitions: int | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        advanced_filter: Filter | dict[str, Any] | None = None,
        sort: SortSpec | list[SortSpec] | None = None,
    ) -> ExtendedTimeSeriesList:
        """`List time series <https://developer.cognite.com/api#tag/Time-series/operation/listTimeSeries>`_

        Args:
            name (str | None): Name of the time series. Often referred to as tag.
            unit (str | None): Unit of the time series.
            unit_external_id (str | None): Filter on unit external ID.
            unit_quantity (str | None): Filter on unit quantity.
            is_string (bool | None): Whether the time series is a string time series.
            is_step (bool | None): Whether the time series is a step (piecewise constant) time series.
            asset_ids (Sequence[int] | None): List time series related to these assets.
            asset_external_ids (SequenceNotStr[str] | None): List time series related to these assets.
            asset_subtree_ids (int | Sequence[int] | None): Only include time series that are related to an asset in a subtree rooted at any of these assetIds. If the total size of the given subtrees exceeds 100,000 assets, an error will be returned.
            asset_subtree_external_ids (str | SequenceNotStr[str] | None): Only include time series that are related to an asset in a subtree rooted at any of these assetExternalIds. If the total size of the given subtrees exceeds 100,000 assets, an error will be returned.
            data_set_ids (int | Sequence[int] | None): Return only time series in the specified data set(s) with this id / these ids.
            data_set_external_ids (str | SequenceNotStr[str] | None): Return only time series in the specified data set(s) with this external id / these external ids.
            metadata (dict[str, Any] | None): Custom, application specific metadata. String key -> String value
            external_id_prefix (str | None): Filter by this (case-sensitive) prefix for the external ID.
            created_time (dict[str, Any] | None):  Range between two timestamps. Possible keys are `min` and `max`, with values given as time stamps in ms.
            last_updated_time (dict[str, Any] | None):  Range between two timestamps. Possible keys are `min` and `max`, with values given as time stamps in ms.
            partitions (int | None): Retrieve resources in parallel using this number of workers (values up to 10 allowed), limit must be set to `None` (or `-1`).
            limit (int | None): Maximum number of time series to return.  Defaults to 25. Set to -1, float("inf") or None to return all items.
            advanced_filter (Filter | dict[str, Any] | None): Advanced filter query using the filter DSL (Domain Specific Language). It allows defining complex filtering expressions that combine simple operations, such as equals, prefix, exists, etc., using boolean operators and, or, and not. See examples below for usage.
            sort (SortSpec | list[SortSpec] | None): The criteria to sort by. Defaults to desc for `_score_` and asc for all other properties. Sort is not allowed if `partitions` is used.

        Returns:
            TimeSeriesList: The requested time series.

        .. note::
            When using `partitions`, there are few considerations to keep in mind:
                * `limit` has to be set to `None` (or `-1`).
                * API may reject requests if you specify more than 10 partitions. When Cognite enforces this behavior, the requests result in a 400 Bad Request status.
                * Partitions are done independently of sorting: there's no guarantee of the sort order between elements from different partitions. For this reason providing a `sort` parameter when using `partitions` is not allowed.

        Examples:

            List time series:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.time_series.list(limit=5)

            Iterate over time series:

                >>> for ts in client.time_series:
                ...     ts # do something with the time series

            Iterate over chunks of time series to reduce memory load:

                >>> for ts_list in client.time_series(chunk_size=2500):
                ...     ts_list # do something with the time series

            Using advanced filter, find all time series that have a metadata key 'timezone' starting with 'Europe',
            and sort by external id ascending:

                >>> from cognite.client.data_classes import filters
                >>> in_timezone = filters.Prefix(["metadata", "timezone"], "Europe")
                >>> res = client.time_series.list(advanced_filter=in_timezone, sort=("external_id", "asc"))

            Note that you can check the API documentation above to see which properties you can filter on
            with which filters.

            To make it easier to avoid spelling mistakes and easier to look up available properties
            for filtering and sorting, you can also use the `TimeSeriesProperty` and `SortableTimeSeriesProperty` Enums.

                >>> from cognite.client.data_classes import filters
                >>> from cognite.client.data_classes.time_series import TimeSeriesProperty, SortableTimeSeriesProperty
                >>> in_timezone = filters.Prefix(TimeSeriesProperty.metadata_key("timezone"), "Europe")
                >>> res = client.time_series.list(
                ...     advanced_filter=in_timezone,
                ...     sort=(SortableTimeSeriesProperty.external_id, "asc"))

            Combine filter and advanced filter:

                >>> from cognite.client.data_classes import filters
                >>> not_instrument_lvl5 = filters.And(
                ...    filters.ContainsAny("labels", ["Level5"]),
                ...    filters.Not(filters.ContainsAny("labels", ["Instrument"]))
                ... )
                >>> res = client.time_series.list(asset_subtree_ids=[123456], advanced_filter=not_instrument_lvl5)
        """
        asset_subtree_ids_processed = process_asset_subtree_ids(asset_subtree_ids, asset_subtree_external_ids)
        data_set_ids_processed = process_data_set_ids(data_set_ids, data_set_external_ids)

        filter = TimeSeriesFilter(
            name=name,
            unit=unit,
            unit_external_id=unit_external_id,
            unit_quantity=unit_quantity,
            is_step=is_step,
            is_string=is_string,
            asset_ids=asset_ids,
            asset_external_ids=asset_external_ids,
            asset_subtree_ids=asset_subtree_ids_processed,
            metadata=metadata,
            data_set_ids=data_set_ids_processed,
            created_time=created_time,
            last_updated_time=last_updated_time,
            external_id_prefix=external_id_prefix,
        ).dump(camel_case=True)

        prep_sort = prepare_filter_sort(sort, TimeSeriesSort)
        self._validate_filter(advanced_filter)

        return self._list(
            list_cls=ExtendedTimeSeriesList,
            resource_cls=ExtendedTimeSeries,
            method="POST",
            filter=filter,
            advanced_filter=advanced_filter,
            sort=prep_sort,
            limit=limit,
            partitions=partitions,
            api_subversion="alpha",
        )



import yaml
from cognite.client.data_classes import TimeSeriesWrite
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.loaders import TimeSeriesLoader
from tests.test_unit.approval_client import ApprovalToolkitClient
from tests.test_unit.approval_client.client import LookUpAPIMock


class TestTimeSeriesLoader:
    timeseries_yaml = """
externalId: pi_160696
name: VAL_23-PT-92504:X.Value
dataSetExternalId: ds_timeseries_oid
isString: false
metadata:
  compdev: '0'
  location5: '2'
isStep: false
description: PH 1stStgSuctCool Gas Out
"""

    def test_load_skip_validation_with_preexisting_dataset(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
    ) -> None:
        loader = TimeSeriesLoader(toolkit_client_approval.mock_client, None)
        ts_dict = yaml.safe_load(self.timeseries_yaml)
        data_set_external_id = ts_dict["dataSetExternalId"]
        expected_id = LookUpAPIMock._create_id(data_set_external_id)

        loaded = loader.load_resource(ts_dict, is_dry_run=False)

        assert loaded.data_set_id == expected_id

    def test_load_skip_validation_no_preexisting_dataset(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
    ) -> None:
        loader = TimeSeriesLoader(toolkit_client_approval.mock_client, None)
        ts_dict = yaml.safe_load(self.timeseries_yaml)

        def id_missing(*args):
            return -1

        toolkit_client_approval.mock_client.lookup.data_sets.id.side_effect = id_missing

        loaded = loader.load_resource(ts_dict, is_dry_run=False)

        assert loaded.data_set_id == -1

    def test_load_timeseries_ref_not_yet_deployed(self) -> None:
        actual_is_dry_run: list[bool] = []

        def mock_id_lookup(
            external_id: str | SequenceNotStr[str], is_dry_run: bool = False, allow_empty: bool = False
        ) -> int | list[int]:
            actual_is_dry_run.append(is_dry_run)
            return -1 if isinstance(external_id, str) else [-1] * len(external_id)

        with monkeypatch_toolkit_client() as client:
            client.lookup.data_sets.id.side_effect = mock_id_lookup
            client.lookup.assets.id.side_effect = mock_id_lookup
            client.lookup.security_categories.id.side_effect = mock_id_lookup
            loader = TimeSeriesLoader.create_loader(client)

        resource = {
            "externalId": "MyTimeseries",
            "assetExternalId": "my_asset",
            "dataSetExternalId": "my_dataset",
            "securityCategoryNames": ["my_security_category"],
        }

        ts = loader.load_resource(resource.copy(), is_dry_run=True)
        _ = loader.load_resource(resource.copy(), is_dry_run=False)

        assert isinstance(ts, TimeSeriesWrite)
        assert ts.asset_id == -1
        assert ts.data_set_id == -1
        assert ts.security_categories == [-1]
        assert actual_is_dry_run == [True] * 3 + [False] * 3



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from cognite.client.data_classes.data_modeling import Container, ContainerProperty, Text

from cognite_toolkit._cdf_tk._parameters import read_parameters_from_dict
from cognite_toolkit._cdf_tk.loaders import ContainerLoader, ResourceWorker
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestContainerLoader:
    @pytest.mark.parametrize(
        "item",
        [
            pytest.param(
                {
                    "properties": {
                        "myDirectRelation": {
                            "name": "my direct relation",
                            "type": {
                                "type": "direct",
                                "container": {
                                    "type": "container",
                                    "space": "sp_my_space",
                                    "externalId": "my_container",
                                },
                            },
                        }
                    }
                },
                id="Direct relation property with require constraint.",
            ),
        ],
    )
    def test_valid_spec(self, item: dict):
        spec = ContainerLoader.get_write_cls_parameter_spec()
        dumped = read_parameters_from_dict(item)

        extra = dumped - spec

        assert not extra, f"Extra keys: {extra}"

    def test_unchanged_used_for_not_set(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        loader = ContainerLoader.create_loader(toolkit_client_approval.mock_client)
        raw_file = """space: sp_enterprise_process_industry_full
externalId: Toolkit360Image
properties:
  UUID:
    type:
      list: false
      collation: ucs_basic
      type: text
    immutable: false
    nullable: true
    autoIncrement: false
constraints: {}
indexes: {}
"""
        file = MagicMock(spec=Path)
        file.read_text.return_value = raw_file
        cdf_container = Container(
            space="sp_enterprise_process_industry_full",
            external_id="Toolkit360Image",
            last_updated_time=1739469813633,
            created_time=1739469813633,
            description=None,
            name=None,
            used_for="node",
            is_global=False,
            properties={"UUID": ContainerProperty(type=Text())},
            indexes={},
            constraints={},
        )

        toolkit_client_approval.append(Container, [cdf_container])

        worker = ResourceWorker(loader)
        to_create, to_change, to_delete, unchanged = worker.prepare_resources([file])
        assert {
            "create": len(to_create),
            "change": len(to_change),
            "delete": len(to_delete),
            "unchanged": len(unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import SecurityCategoriesYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_security_category_test_cases() -> Iterable:
    yield pytest.param(
        {"externalId": "MyCategory"},
        {"Missing required field: 'name'", "Unused field: 'externalId'"},
        id="Missing required field: name",
    )


class TestSecuritytCategoriesYAML:
    @pytest.mark.parametrize("data", list(find_resources("SecurityCategory")))
    def test_load_valid_security_categories(self, data: dict[str, object]) -> None:
        loaded = SecurityCategoriesYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_security_category_test_cases()))
    def test_invalid_asset_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, SecurityCategoriesYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import GroupYAML, TimeSeriesYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import as_json_path, validate_resource_yaml_pydantic


def timeseries_yaml_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "my_timeseries"},
        [
            "Missing required field: 'externalId'",
        ],
        id="Missing externalId",
    )
    yield pytest.param(
        {"externalId": "my_timeseries", "type": "numeric"},
        [
            "Unused field: 'type'",
        ],
        id="Unused field type",
    )

    yield pytest.param(
        [
            {"externalId": "my_timeseries", "name": "my_timeseries"},
            {"name": "my_timeseries_2"},
        ],
        [
            "In item [2] missing required field: 'externalId'",
        ],
        id="Invalid second element of list of timeseries",
    )

    yield pytest.param(
        [
            {"externalId": "my_timeseries", "nam": "my_timeseries"},
            {"name": "my_timeseries_2", "type": "numeric"},
        ],
        [
            "In item [1] unused field: 'nam'",
            "In item [2] missing required field: 'externalId'",
            "In item [2] unused field: 'type'",
        ],
        id="Multiple issues in a list of timeseries",
    )


def group_yaml_test_cases() -> Iterable:
    yield pytest.param(
        {"sourceId": "123-345"},
        {"Missing required field: 'name'"},
        id="Missing name",
    )
    yield pytest.param(
        {
            "name": "invalid-group",
            "capabilities": [{"dataModelInstancesAcl": {"actions": ["INVALID_ACTION"]}}],
            "members": "allUserAccounts",
        },
        {
            "In capabilities[1].actions input should be 'READ', 'WRITE' or 'WRITE_PROPERTIES'. Got 'INVALID_ACTION'.",
            "In capabilities[1] missing required field: 'scope'",
        },
        id="Invalid action and missing scope",
    )

    yield pytest.param(
        [
            {
                "name": "invalid-group",
                "capabilities": [
                    {"dataModelsAcl": {"actions": ["WRITE"], "scope": {"notExisting": {"spaceIds": {"my_space"}}}}}
                ],
                "members": "allUserAccounts",
            }
        ],
        {
            "In item [1].capabilities[1] invalid scope name 'notExisting'. Expected all or spaceIdScope",
        },
        id="Invalid scope name",
    )


class TestValidateResourceYAML:
    @pytest.mark.parametrize("data, expected_errors", list(timeseries_yaml_test_cases()))
    def test_validate_timeseries_resource_yaml(self, data: dict | list, expected_errors: list[str]) -> None:
        """Test the validate_resource_yaml function for TimeSeriesYAML."""
        warning_list = validate_resource_yaml_pydantic(data, TimeSeriesYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert format_warning.errors == tuple(expected_errors)

    @pytest.mark.parametrize("data, expected_errors", list(group_yaml_test_cases()))
    def test_validate_group_resource_yaml(self, data: dict | list, expected_errors: set[str]) -> None:
        """Test the validate_resource_yaml function for GroupYAML."""
        warning_list = validate_resource_yaml_pydantic(data, GroupYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors


class TestAsJsonPath:
    @pytest.mark.parametrize(
        "loc, expected",
        [
            (("a", "b", "c"), "a.b.c"),
            (("a", 1, "c"), "a[2].c"),
            (("a", 1, 2), "a[2][3]"),
            (("a",), "a"),
            ((), ""),
            ((1,), "item [2]"),
        ],
    )
    def test_as_json_path(self, loc: tuple[str | int, ...], expected: str) -> None:
        assert as_json_path(loc) == expected



from __future__ import annotations

import inspect
import re
from pathlib import Path
from typing import Any

from cognite.client.data_classes._base import CogniteObject
from cognite.client.utils._text import to_camel_case, to_snake_case
from pydantic import TypeAdapter, ValidationError
from pydantic_core import ErrorDetails

from cognite_toolkit._cdf_tk._parameters import ParameterSpecSet, read_parameters_from_dict
from cognite_toolkit._cdf_tk.data_classes import BuildVariables
from cognite_toolkit._cdf_tk.loaders import NodeLoader
from cognite_toolkit._cdf_tk.resource_classes import BaseModelResource
from cognite_toolkit._cdf_tk.tk_warnings import (
    CaseTypoWarning,
    DataSetMissingWarning,
    MissingRequiredParameterWarning,
    TemplateVariableWarning,
    UnusedParameterWarning,
    WarningList,
)

__all__ = ["validate_data_set_is_set", "validate_modules_variables", "validate_resource_yaml"]

from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning


def validate_modules_variables(variables: BuildVariables, filepath: Path) -> WarningList:
    """Checks whether the config file has any issues.

    Currently, this checks for:
        * Non-replaced template variables, such as <change_me>.

    Args:
        variables: The variables to check.
        filepath: The filepath of the config.yaml.
    """
    warning_list: WarningList = WarningList()
    pattern = re.compile(r"<.*?>")
    for variable in variables:
        if isinstance(variable.value, str) and pattern.match(variable.value):
            warning_list.append(
                TemplateVariableWarning(filepath, variable.value, variable.key, ".".join(variable.location.parts))
            )
    return warning_list


def validate_data_set_is_set(
    raw: dict[str, Any] | list[dict[str, Any]],
    resource_cls: type[CogniteObject],
    filepath: Path,
    identifier_key: str = "externalId",
) -> WarningList:
    warning_list: WarningList = WarningList()
    signature = inspect.signature(resource_cls.__init__)
    if "data_set_id" not in set(signature.parameters.keys()):
        return warning_list

    if isinstance(raw, list):
        for item in raw:
            warning_list.extend(validate_data_set_is_set(item, resource_cls, filepath, identifier_key))
        return warning_list

    if "dataSetExternalId" in raw or "dataSetId" in raw:
        return warning_list

    value = raw.get(identifier_key, raw.get(to_snake_case(identifier_key), f"No identifier {identifier_key}"))
    warning_list.append(DataSetMissingWarning(filepath, value, identifier_key, resource_cls.__name__))
    return warning_list


def validate_resource_yaml(
    data: dict | list, spec: ParameterSpecSet, source_file: Path, element: int | None = None
) -> WarningList:
    if spec.spec_name == NodeLoader.__name__:
        # Special case for NodeLoader as it has options for API call parameters
        if isinstance(data, list):
            return _validate_resource_yaml(data, spec, source_file)
        elif isinstance(data, dict) and "node" in data:
            return _validate_resource_yaml(data["node"], spec, source_file)
        elif isinstance(data, dict) and "nodes" in data:
            return _validate_resource_yaml(data["nodes"], spec, source_file)
        else:
            return _validate_resource_yaml(data, spec, source_file)
    else:
        return _validate_resource_yaml(data, spec, source_file, element)


def _validate_resource_yaml(
    data: dict | list, spec: ParameterSpecSet, source_file: Path, element: int | None = None
) -> WarningList:
    warnings: WarningList = WarningList()
    if isinstance(data, list):
        for no, item in enumerate(data, 1):
            warnings.extend(_validate_resource_yaml(item, spec, source_file, no))
        return warnings
    elif not isinstance(data, dict):
        raise NotImplementedError("Note: This function only supports top-level and lists dictionaries.")

    actual_parameters = read_parameters_from_dict(data)
    unused_parameters = actual_parameters - spec
    unused_cased = unused_parameters.as_camel_case() - spec
    typo_parameters = unused_parameters - unused_cased
    for parameter in typo_parameters:
        key = parameter.key
        warnings.append(CaseTypoWarning(source_file, element, parameter.path, key, to_camel_case(key)))

    unused_parameters = unused_parameters - typo_parameters
    for parameter in unused_parameters:
        key = parameter.key
        warnings.append(UnusedParameterWarning(source_file, element, parameter.path, key))

    # Only checking the top level for now. This can be expanded to check nested parameters.
    missing = spec.required(level=1) - actual_parameters
    for spec_param in missing:
        warnings.append(MissingRequiredParameterWarning(source_file, element, spec_param.path, spec_param.key))

    return warnings


def validate_resource_yaml_pydantic(
    data: dict[str, object] | list[dict[str, object]], validation_cls: type[BaseModelResource], source_file: Path
) -> WarningList:
    """Validates the resource given as a dictionary or list of dictionaries with the given pydantic model.

    Args:
        data: The data to validate.
        validation_cls: The pydantic model to validate against.
        source_file: The source file of the resource.

    Returns:
        A list of warnings.

    """
    warning_list: WarningList = WarningList()
    if isinstance(data, dict):
        try:
            validation_cls.model_validate(data, strict=True)
        except ValidationError as e:
            warning_list.append(ResourceFormatWarning(source_file, tuple(_humanize_validation_error(e))))
    elif isinstance(data, list):
        try:
            TypeAdapter(list[validation_cls]).validate_python(data)  # type: ignore[valid-type]
        except ValidationError as e:
            warning_list.append(ResourceFormatWarning(source_file, tuple(_humanize_validation_error(e))))
    else:
        raise ValueError(f"Expected a dictionary or list of dictionaries, got {type(data)}.")
    return warning_list


def _humanize_validation_error(error: ValidationError) -> list[str]:
    """Converts a ValidationError to a human-readable format.

    This overwrites the default error messages from Pydantic to be better suited for Toolkit users.

    Args:
        error: The ValidationError to convert.

    Returns:
        A list of human-readable error messages.
    """
    errors: list[str] = []
    item: ErrorDetails
    for item in error.errors(include_input=True, include_url=False):
        loc = item["loc"]
        error_type = item["type"]
        if error_type == "missing":
            msg = f"Missing required field: {loc[-1]!r}"
        elif error_type == "extra_forbidden":
            msg = f"Unused field: {loc[-1]!r}"
        elif error_type == "value_error":
            msg = str(item["ctx"]["error"])
        elif error_type in {"literal_error", "list_type"}:
            msg = f"{item['msg']}. Got {item['input']!r}."
        elif error_type == "string_type" and len(loc) >= 2 and loc[-2] == "metadata":
            key = loc[-1]
            msg = f"The key {key!r} should be a valid string. Got {item['input']!r} of type {type(item['input']).__name__}. Hint: Use double quotes to force string."
        elif error_type == "string_type":
            msg = f"{item['msg']}. Got {item['input']!r} of type {type(item['input']).__name__}. Hint: Use double quotes to force string."
        elif error_type == "model_type":
            model_name = item["ctx"].get("class_name", "unknown")
            msg = f"Input must be an object of type {model_name}. Got {item['input']!r} of type {type(item['input']).__name__}."
        elif error_type in {
            "int_type",
            "bool_type",
            "datetime_type",
            "decimal_type",
            "float_type",
            "time_type",
            "timedelta_type",
        }:
            msg = f"{item['msg']}. Got {item['input']!r} of type {type(item['input']).__name__}."
        else:
            # Default to the Pydantic error message
            msg = item["msg"]
        if len(loc) > 1:
            msg = f"In {as_json_path(loc[:-1])} {msg[0].casefold()}{msg[1:]}"
        elif len(loc) == 1 and isinstance(loc[0], str) and error_type not in {"extra_forbidden", "missing"}:
            msg = f"In field {loc[0]} {msg[0].casefold()}{msg[1:]}"
        errors.append(msg)
    return errors


def as_json_path(loc: tuple[str | int, ...]) -> str:
    """Converts a location tuple to a JSON path.

    Args:
        loc: The location tuple to convert.

    Returns:
        A JSON path string.
    """
    if not loc:
        return ""
    # +1 to convert from 0-based to 1-based indexing
    prefix = ""
    if isinstance(loc[0], int):
        prefix = "item "

    suffix = ".".join([str(x) if isinstance(x, str) else f"[{x + 1}]" for x in loc]).replace(".[", "[")
    return f"{prefix}{suffix}"



from __future__ import annotations

import pytest

from cognite_toolkit._cdf_tk.data_classes import Environment
from tests.data import PROJECT_FOR_TEST


@pytest.fixture(scope="session")
def project_for_test_config_dev_yaml() -> str:
    return (PROJECT_FOR_TEST / "config.dev.yaml").read_text()


@pytest.fixture(scope="session")
def dummy_environment() -> Environment:
    return Environment(
        name="dev",
        project="my_project",
        validation_type="dev",
        selected=["none"],
    )



from typing import Any

import typer
from rich import print

from cognite_toolkit._cdf_tk.commands import ProfileCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class ProfileApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("asset-centric")(self.asset_centric)

    def main(self, ctx: typer.Context) -> None:
        """Commands profile functionality"""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf profile --help[/] for more information.")

    @staticmethod
    def asset_centric(
        ctx: typer.Context,
        verbose: bool = False,
    ) -> None:
        """This command gives an overview over the metadata and labels for each of the asset-centric resources.
        This shows an approximation of unstructured data count. This can, for example, be used to estimate the
        effort to model this data in data modeling."""
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = ProfileCommand()
        cmd.run(
            lambda: cmd.asset_centric(
                client,
                verbose,
            )
        )



from pathlib import Path
from typing import Annotated, Any

import typer

from cognite_toolkit._cdf_tk.commands import MigrateTimeseriesCommand, MigrationPrepareCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class MigrateApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("prepare")(self.prepare)
        self.command("timeseries")(self.timeseries)

    def main(self, ctx: typer.Context) -> None:
        """Migrate resources from Asset-Centric to data modeling in CDF."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf migrate --help[/] for more information.")

    @staticmethod
    def prepare(
        ctx: typer.Context,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="The preparation will not be executed, only report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Prepare the migration of resources from Asset-Centric to data modeling in CDF.

        This means deploying the CogniteMigration data model that contains the Mapping view. This will be used
        to store the mapping from Asset-Centric resources to the new data modeling resources.

        This mapping will be used when migrating applications such as Canvas, Charts, as well as resources that
        depend on the primary resources 3D and annotations.
        """
        client = EnvironmentVariables.create_from_environment().get_client(enable_set_pending_ids=True)
        cmd = MigrationPrepareCommand()
        cmd.run(
            lambda: cmd.deploy_cognite_migration(
                client,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def timeseries(
        ctx: typer.Context,
        mapping_file: Annotated[
            Path,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from TimeSeries to CogniteTimeSeries. "
                "This file is expected to have the following columns: [id/externalId, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ],
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate TimeSeries to CogniteTimeSeries."""

        client = EnvironmentVariables.create_from_environment().get_client(enable_set_pending_ids=True)
        cmd = MigrateTimeseriesCommand()
        cmd.run(
            lambda: cmd.migrate_timeseries(
                client,
                mapping_file=mapping_file,
                dry_run=dry_run,
                verbose=verbose,
            )
        )



from typing import Annotated, Any, Optional

import typer
from rich import print

from cognite_toolkit._cdf_tk.commands import PurgeCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class PurgeApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("dataset")(self.purge_dataset)
        self.command("space")(self.purge_space)

    def main(self, ctx: typer.Context) -> None:
        """Commands purge functionality"""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf purge --help[/] for more information.")

    def purge_dataset(
        self,
        ctx: typer.Context,
        external_id: Annotated[
            Optional[str],
            typer.Argument(
                help="External id of the dataset to purge. If not provided, interactive mode will be used.",
            ),
        ] = None,
        include_dataset: Annotated[
            bool,
            typer.Option(
                "--include-dataset",
                "-i",
                help="Include dataset in the purge. This will also archive the dataset.",
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run, do dry-run if present.",
            ),
        ] = False,
        auto_yes: Annotated[
            bool,
            typer.Option(
                "--yes",
                "-y",
                help="Automatically confirm that you are sure you want to purge the dataset.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will delete the contents of the specified dataset"""
        cmd = PurgeCommand()
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd.run(
            lambda: cmd.dataset(
                client,
                external_id,
                include_dataset,
                dry_run,
                auto_yes,
                verbose,
            )
        )

    def purge_space(
        self,
        ctx: typer.Context,
        space: Annotated[
            Optional[str],
            typer.Argument(
                help="Space to purge. If not provided, interactive mode will be used.",
            ),
        ] = None,
        include_space: Annotated[
            bool,
            typer.Option(
                "--include-space",
                "-i",
                help="Include space in the purge. This will also delete the space.",
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run, do dry-run if present.",
            ),
        ] = False,
        auto_yes: Annotated[
            bool,
            typer.Option(
                "--yes",
                "-y",
                help="Automatically confirm that you are sure you want to purge the space.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will delete the contents of the specified space."""

        cmd = PurgeCommand()
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd.run(
            lambda: cmd.space(
                client,
                space,
                include_space,
                dry_run,
                auto_yes,
                verbose,
            )
        )



from __future__ import annotations

import ast
import datetime
import json
import os
import platform
import re
import shutil
import textwrap
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, cast

import questionary
from cognite.client._api.functions import ALLOWED_HANDLE_ARGS
from cognite.client.credentials import OAuthClientCredentials, OAuthInteractive, Token
from cognite.client.data_classes import ClientCredentials, WorkflowTriggerUpsert
from cognite.client.data_classes.transformations import TransformationList
from cognite.client.data_classes.transformations.common import NonceCredentials
from cognite.client.data_classes.workflows import (
    FunctionTaskParameters,
    WorkflowExecutionDetailed,
    WorkflowVersionId,
    WorkflowVersionUpsert,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils import ms_to_datetime
from rich import print
from rich.progress import Progress
from rich.table import Table

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.functions import FunctionScheduleID
from cognite_toolkit._cdf_tk.constants import _RUNNING_IN_BROWSER
from cognite_toolkit._cdf_tk.data_classes import BuiltResourceFull, ModuleResources
from cognite_toolkit._cdf_tk.exceptions import (
    AuthorizationError,
    ToolkitFileNotFoundError,
    ToolkitInvalidFunctionError,
    ToolkitMissingResourceError,
    ToolkitNotADirectoryError,
    ToolkitNotSupported,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.hints import verify_module_directory
from cognite_toolkit._cdf_tk.loaders import FunctionLoader, FunctionScheduleLoader, WorkflowVersionLoader
from cognite_toolkit._cdf_tk.loaders._resource_loaders.workflow_loaders import WorkflowTriggerLoader
from cognite_toolkit._cdf_tk.tk_warnings import MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import in_dict
from cognite_toolkit._cdf_tk.utils.auth import CLIENT_NAME, EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.file import safe_read, safe_rmtree, safe_write

from ._base import ToolkitCommand


@dataclass
class FunctionCallArgs:
    data: dict[str, Any] = field(default_factory=dict)
    authentication: ClientCredentials | None = None
    client_id_env_name: str | None = None
    client_secret_env_name: str | None = None


class RunFunctionCommand(ToolkitCommand):
    virtual_env_folder = "function_local_venvs"
    default_readme_md = """# Local Function Quality Assurance

This directory contains virtual environments for running functions locally. This is
intended to test the function before deploying it to CDF or to debug issues with a deployed function.

"""
    import_check_py = """import sys
from pathlib import Path

# This is necessary to import adjacent modules in the function code.
sys.path.insert(0, str(Path(__file__).parent / "local_code"))

from local_code.{handler_import} import handle # noqa: E402


def main() -> None:
    print("Imported function successfully: " + handle.__name__)


if __name__ == "__main__":
    main()

"""
    run_check_py = """import os
import sys

from pathlib import Path
from pprint import pprint

from cognite.client import CogniteClient, ClientConfig
from cognite.client.credentials import {credentials_cls}

# This is necessary to import adjacent modules in the function code.
sys.path.insert(0, str(Path(__file__).parent / "local_code"))

from local_code.{handler_import} import handle # noqa: E402
{load_dotenv}

def main() -> None:
    credentials = {credentials_cls}(
        {credentials_args}
    )

    client = CogniteClient(
        config=ClientConfig(
            client_name="{client_name}",
            project="{project}",
            base_url="{base_url}",
            credentials=credentials,
        )
    )

    print("{function_external_id} LOGS:")
    response = handle(
        {handler_args}
    )

    print("{function_external_id} RESPONSE:")
    pprint(response)


if __name__ == "__main__":
    main()
"""

    def run_cdf(
        self,
        env_vars: EnvironmentVariables,
        organization_dir: Path,
        build_env_name: str | None,
        external_id: str | None = None,
        data_source: str | WorkflowVersionId | None = None,
        wait: bool = False,
    ) -> bool:
        if organization_dir in {Path("."), Path("./")}:
            organization_dir = Path.cwd()
        verify_module_directory(organization_dir, build_env_name)

        resources = ModuleResources(organization_dir, build_env_name)
        is_interactive = external_id is None
        external_id = self._get_function(external_id, resources).identifier
        call_args = self._get_call_args(data_source, external_id, resources, env_vars.dump(), is_interactive)
        client = env_vars.get_client()
        function = client.functions.retrieve(external_id=external_id)
        if function is None:
            raise ToolkitMissingResourceError(
                f"Could not find function with external id {external_id}. Have you deployed it?"
            )

        if is_interactive:
            wait = questionary.confirm("Do you want to wait for the function to complete?").ask()

        # Todo: Get one shot token using the call_args.authentication
        session = client.iam.sessions.create(session_type="ONESHOT_TOKEN_EXCHANGE")
        result = client.functions.call(external_id=external_id, data=call_args.data, wait=False, nonce=session.nonce)

        table = Table(title=f"Function {external_id!r}, id {function.id!r}")
        table.add_column("Info", justify="left")
        table.add_column("Value", justify="left", style="green")
        table.add_row("Call id", str(result.id))
        table.add_row("Status", str(result.status))
        table.add_row("Created time", str(ms_to_datetime(result.start_time)))
        print(table)

        if not wait:
            return True

        max_time = client.functions.limits().timeout_minutes * 60
        with Progress() as progress:
            call_task = progress.add_task("Waiting for function call to complete...", total=max_time)
            start_time = time.time()
            duration = 0.0
            sleep_time = 1
            while result.status.casefold() == "running" and duration < max_time:
                time.sleep(sleep_time)
                sleep_time = min(sleep_time * 2, 60)
                result.update()
                duration = time.time() - start_time
                progress.advance(call_task, advance=duration)
            progress.advance(call_task, advance=max_time - duration)
            progress.stop()
        table = Table(title=f"Function {external_id}, id {function.id}")
        table.add_column("Info", justify="left")
        table.add_column("Value", justify="left", style="green")
        table.add_row("Call id", str(result.id))
        table.add_row("Status", str(result.status))
        created_time = ms_to_datetime(result.start_time)
        finished_time = ms_to_datetime(result.end_time or (datetime.datetime.now().timestamp() * 1000))
        table.add_row("Created time", str(created_time))
        table.add_row("Finished time", str(finished_time))
        run_time = finished_time - created_time
        table.add_row("Duration", f"{run_time.total_seconds():,} seconds")
        if result.error is not None:
            table.add_row("Error", str(result.error.get("message", "Empty error")))
            table.add_row("Error trace", str(result.error.get("trace", "Empty trace")))
        response = client.functions.calls.get_response(call_id=result.id or 0, function_id=function.id)
        table.add_row("Result", str(json.dumps(response, indent=2, sort_keys=True)))
        logs = client.functions.calls.get_logs(call_id=result.id or 0, function_id=function.id)
        table.add_row("Logs", str(logs))
        print(table)
        return True

    @staticmethod
    def _get_function(external_id: str | None, resources: ModuleResources) -> BuiltResourceFull[str]:
        function_builds_by_identifier = {
            build.identifier: build for build in resources.list_resources(str, "functions", FunctionLoader.kind)
        }

        if len(function_builds_by_identifier) == 0:
            raise ToolkitMissingResourceError(
                "No functions found in modules. Suggest running `cdf modules list` to verify."
            )

        if external_id is None:
            # Interactive mode
            external_id = questionary.select(
                "Select function to run", choices=list(function_builds_by_identifier.keys())
            ).ask()
        elif external_id not in function_builds_by_identifier.keys():
            raise ToolkitMissingResourceError(f"Could not find function with external id {external_id}")
        return function_builds_by_identifier[external_id]

    @classmethod
    def _get_call_args(
        cls,
        data_source: str | WorkflowVersionId | None,
        function_external_id: str,
        resources: ModuleResources,
        environment_variables: dict[str, str | None],
        is_interactive: bool,
    ) -> FunctionCallArgs:
        if data_source is None and (
            not is_interactive or not questionary.confirm("Do you want to provide input data for the function?").ask()
        ):
            return FunctionCallArgs()
        if is_interactive:
            data, credentials = cls._get_call_args_interactive(function_external_id, resources)
        elif data_source is not None:
            data, credentials = cls._geta_call_args_from_data_source(data_source, function_external_id, resources)
        else:
            raise ToolkitValueError("Data source is required when not in interactive mode.")

        if credentials is None:
            return FunctionCallArgs(data)

        if cls._is_environ_var(credentials.client_id):
            env_var = cls._clean_environ_var(credentials.client_id)
            if not (client_id_value := environment_variables.get(env_var)):
                raise ToolkitValueError(f"Missing environment variable {env_var} for function client ID")
            client_id_name = env_var
        else:
            client_id_name = None
            client_id_value = credentials.client_id

        if cls._is_environ_var(credentials.client_secret):
            env_var = cls._clean_environ_var(credentials.client_secret)
            if not (client_secret_value := environment_variables.get(env_var)):
                raise ToolkitValueError(f"Missing environment variable {env_var} for function client secret")
            client_secret_name = env_var
        else:
            client_secret_name = None
            client_secret_value = credentials.client_secret

        return FunctionCallArgs(
            data, ClientCredentials(client_id_value, client_secret_value), client_id_name, client_secret_name
        )

    @staticmethod
    def _is_environ_var(value: str) -> bool:
        return value.startswith("${") and value.endswith("}")

    @staticmethod
    def _clean_environ_var(value: str) -> str:
        return value.removeprefix("${").removesuffix("}")

    @staticmethod
    def _get_call_args_interactive(
        function_external_id: str, resources: ModuleResources
    ) -> tuple[dict[str, Any], ClientCredentials | None]:
        schedules = resources.list_resources(FunctionScheduleID, "functions", FunctionScheduleLoader.kind)
        options: dict[str, Any] = {
            f"FunctionSchedule: {schedule.identifier.name}": schedule
            for schedule in schedules
            if schedule.identifier.function_external_id == function_external_id
        }
        workflows = resources.list_resources(WorkflowVersionId, "workflows", WorkflowVersionLoader.kind)
        raw_trigger_by_workflow_id: dict[WorkflowVersionId, dict[str, Any]] = {}
        for trigger in resources.list_resources(str, "workflows", WorkflowTriggerLoader.kind):
            raw_trigger = trigger.load_resource_dict({}, validate=False)
            loaded_trigger = WorkflowTriggerUpsert.load(raw_trigger)
            raw_trigger_by_workflow_id[
                WorkflowVersionId(loaded_trigger.workflow_external_id, loaded_trigger.workflow_version)
            ] = raw_trigger

        for workflow in workflows:
            raw_workflow = workflow.load_resource_dict({}, validate=False)
            loaded = WorkflowVersionUpsert.load(raw_workflow)
            for task in loaded.workflow_definition.tasks:
                if (
                    isinstance(task.parameters, FunctionTaskParameters)
                    and task.parameters.external_id == function_external_id
                ):
                    data = task.parameters.data if isinstance(task.parameters.data, dict) else {}
                    raw_trigger = raw_trigger_by_workflow_id.get(workflow.identifier, {})
                    options[f"Workflow: {workflow.identifier.workflow_external_id}"] = (
                        data,
                        raw_trigger.get("authentication"),
                    )

        if len(options) == 0:
            print(f"No schedules or workflows found for this {function_external_id} function.")
            return {}, None
        selected_name: str = questionary.select("Select schedule to run", choices=options).ask()  # type: ignore[arg-type]
        selected = options[selected_name]
        if isinstance(selected, BuiltResourceFull):
            # Schedule
            raw_schedule = selected.load_resource_dict({}, validate=False)
            return raw_schedule.get("data", {}), ClientCredentials.load(
                raw_schedule["authentication"]
            ) if "authentication" in raw_schedule else None
        elif (
            isinstance(selected, tuple)
            and len(selected) == 2
            and isinstance(selected[0], dict)
            and isinstance(selected[1], dict)
        ):
            return selected[0], ClientCredentials.load(selected[1]) if in_dict(
                ["clientId", "clientSecret"], selected[1]
            ) else None
        else:
            raise ToolkitValueError(f"Selected value {selected} is not a valid schedule or workflow.")

    @staticmethod
    def _geta_call_args_from_data_source(
        data_source: str | WorkflowVersionId, function_external_id: str, resources: ModuleResources
    ) -> tuple[dict[str, Any], ClientCredentials | None]:
        data: dict[str, Any] | None = None
        credentials: ClientCredentials | None = None
        workflows = resources.list_resources(WorkflowVersionId, "workflows", WorkflowVersionLoader.kind)
        found = False
        for workflow in workflows:
            if (isinstance(data_source, str) and workflow.identifier.workflow_external_id == data_source) or (
                isinstance(data_source, WorkflowVersionId) and workflow.identifier == data_source
            ):
                raw_workflow = workflow.load_resource_dict({}, validate=False)
                loaded = WorkflowVersionUpsert.load(raw_workflow)
                for task in loaded.workflow_definition.tasks:
                    if (
                        isinstance(task.parameters, FunctionTaskParameters)
                        and task.parameters.external_id == function_external_id
                    ):
                        data = task.parameters.data if isinstance(task.parameters.data, dict) else {}
                        found = True
                        break
            for trigger in resources.list_resources(str, "workflows", WorkflowTriggerLoader.kind):
                raw_trigger = trigger.load_resource_dict({}, validate=False)
                loaded_trigger = WorkflowTriggerUpsert.load(raw_trigger)
                if (isinstance(data_source, str) and loaded_trigger.workflow_external_id == data_source) or (
                    isinstance(data_source, WorkflowVersionId)
                    and WorkflowVersionId(loaded_trigger.workflow_external_id, loaded_trigger.workflow_version)
                    == data_source
                ):
                    if "authentication" in raw_trigger:
                        try:
                            credentials = ClientCredentials.load(raw_trigger["authentication"])
                        except KeyError:
                            ...
                        else:
                            found = True
                    break
            if found:
                return data or {}, credentials

        if not isinstance(data_source, str):
            raise ToolkitValueError(f"Data source {data_source} is not a valid workflow external id.")

        for schedule in resources.list_resources(FunctionScheduleID, "functions", FunctionScheduleLoader.kind):
            if (
                schedule.identifier.function_external_id == function_external_id
                and schedule.identifier.name == data_source
            ):
                raw_schedule = schedule.load_resource_dict({}, validate=False)
                return raw_schedule.get("data", {}), ClientCredentials.load(
                    raw_schedule["authentication"]
                ) if "authentication" in raw_schedule else None
        raise ToolkitMissingResourceError(f"Could not find data for source {data_source}")

    def run_local(
        self,
        env_vars: EnvironmentVariables,
        organization_dir: Path,
        build_env_name: str | None,
        external_id: str | None = None,
        data_source: str | WorkflowVersionId | None = None,
        rebuild_env: bool = False,
        virtual_env_folder_name: str = virtual_env_folder,
    ) -> None:
        try:
            from ._virtual_env import FunctionVirtualEnvironment
        except ImportError:
            if _RUNNING_IN_BROWSER:
                raise ToolkitNotSupported("This functionality is not supported in a browser environment.")
            raise

        if organization_dir in {Path("."), Path("./")}:
            organization_dir = Path.cwd()
        verify_module_directory(organization_dir, build_env_name)

        resources = ModuleResources(organization_dir, build_env_name)
        function_build = self._get_function(external_id, resources)

        function_external_id = function_build.identifier

        virtual_envs_dir = organization_dir / virtual_env_folder_name
        virtual_envs_dir.mkdir(exist_ok=True)
        readme_overview = virtual_envs_dir / "README.md"
        if not readme_overview.exists():
            safe_write(readme_overview, self.default_readme_md)

        function_venv = Path(virtual_envs_dir) / function_external_id
        function_source_code = function_build.source.path.parent / function_external_id
        if not function_source_code.exists():
            raise ToolkitNotADirectoryError(
                f"Could not find function code for {function_external_id}. Expected at {function_source_code.as_posix()}"
            )

        requirements_txt: Path | str = function_source_code / "requirements.txt"
        if not cast(Path, requirements_txt).exists():
            self.warn(
                MediumSeverityWarning(
                    "No requirements.txt found for function, "
                    "it is recommended that you have one and pin your "
                    "dependencies including the cognite-sdk"
                )
            )

            # Default to install only the SDK in the latest version
            requirements_txt = "cognite-sdk\n"

        print(f"Setting up virtual environment for function {function_external_id}...")
        # Todo: Ensure cognite-sdk is installed in the virtual environment?
        virtual_env = FunctionVirtualEnvironment(requirements_txt, rebuild_env)

        virtual_env.create(function_venv / ".venv")

        print(
            f"    [green]✓ 1/4[/green] Function {function_external_id!r} virtual environment setup complete: "
            f"'requirements.txt' file is valid."
        )

        function_destination_code = function_venv / "local_code"
        if function_destination_code.exists():
            safe_rmtree(function_destination_code)
        shutil.copytree(function_source_code, function_destination_code)
        init_py = function_destination_code / "__init__.py"
        if not init_py.exists():
            # We need a __init__ to avoid import errors when running the function code.
            init_py.touch()

        function_dict = function_build.load_resource_dict(env_vars.dump(), validate=False)
        handler_file = function_dict.get("functionPath", "handler.py")
        handler_path = function_destination_code / handler_file
        if not handler_path.exists():
            raise ToolkitFileNotFoundError("Could not find handler.py file for function.")

        args = self._get_function_args(handler_path, function_name="handle")
        expected = ALLOWED_HANDLE_ARGS
        if not args <= ALLOWED_HANDLE_ARGS:
            raise ToolkitInvalidFunctionError(
                f"Function handle function should have arguments {expected}, got {list(args)}"
            )
        print(
            f"    [green]✓ 2/4[/green] Function {function_external_id!r} the {handler_file!r} "
            f"is valid with arguments {list(args)}."
        )

        import_check = "import_check.py"
        safe_write(
            function_venv / import_check,
            self.import_check_py.format(handler_import=self._create_handler_import(handler_file)),
        )

        virtual_env.execute(Path(import_check), f"import_check {function_external_id}")

        print(f"    [green]✓ 3/4[/green] Function {function_external_id!r} successfully imported code.")

        is_interactive = external_id is None
        call_args = self._get_call_args(
            data_source,
            function_build.identifier,
            resources,
            env_vars.dump(),
            is_interactive,
        )

        run_check_py, env = self._create_run_check_file_with_env(
            env_vars,
            args,
            function_dict,
            function_external_id,
            handler_file,
            call_args,
            function_venv,
        )
        if platform.system() == "Windows":
            if system_root := os.environ.get("SYSTEMROOT"):
                # This is necessary to run python on Windows.
                # https://stackoverflow.com/questions/78652758/cryptic-oserror-winerror-10106-the-requested-service-provider-could-not-be-l
                env["SYSTEMROOT"] = system_root
            # In addition, we need to convert all values to strings.
            env = {k: str(v) for k, v in env.items()}

        run_check = "run_check.py"
        safe_write(function_venv / run_check, run_check_py)
        virtual_env.execute(Path(run_check), f"run_check {function_external_id}", env)

        print(f"    [green]✓ 4/4[/green] Function {function_external_id!r} successfully executed code.")

    def _create_run_check_file_with_env(
        self,
        env_vars: EnvironmentVariables,
        args: set[str],
        function_dict: dict[str, Any],
        function_external_id: str,
        handler_file: str,
        call_args: FunctionCallArgs,
        function_venv_dir: Path,
    ) -> tuple[str, dict[str, str]]:
        if authentication := call_args.authentication:
            if authentication.client_id.startswith("${"):
                raise ToolkitInvalidFunctionError(
                    f"Missing environment variable for clientId in schedule authentication, {authentication.client_id}"
                )
            if authentication.client_secret.startswith("${"):
                raise ToolkitInvalidFunctionError(
                    f"Missing environment variable for clientSecret in schedule authentication, {authentication.client_secret}"
                )
            print(f"Using schedule authentication to run {function_external_id!r}.")
            secret_env_name = call_args.client_secret_env_name or "IDP_CLIENT_SECRET"
            credentials_args = {
                "token_url": f'"{env_vars.idp_token_url}"',
                "client_id": f'"{authentication.client_id}"',
                "client_secret": f'os.environ["{secret_env_name}"]',
                "scopes": str(env_vars.idp_scopes),
            }
            env = {
                secret_env_name: authentication.client_secret,
            }
            credentials_cls = OAuthClientCredentials.__name__
        else:
            print(f"Using Toolkit authentication to run {function_external_id!r}.")
            env = {}
            if env_vars.LOGIN_FLOW == "token":
                credentials_cls = Token.__name__
                credentials_args = {"token": 'os.environ["CDF_TOKEN"]'}
                env["CDF_TOKEN"] = env_vars.CDF_TOKEN  # type: ignore[assignment]
            elif env_vars.LOGIN_FLOW == "interactive":
                credentials_cls = OAuthInteractive.__name__
                credentials_args = {
                    "authority_url": f'"{env_vars.idp_authority_url}"',
                    "client_id": f'"{env_vars.IDP_CLIENT_ID}"',
                    "scopes": str(env_vars.idp_scopes),
                }
            elif env_vars.LOGIN_FLOW == "client_credentials":
                credentials_cls = OAuthClientCredentials.__name__
                credentials_args = {
                    "token_url": f'"{env_vars.idp_token_url}"',
                    "client_id": f'"{env_vars.IDP_CLIENT_ID}"',
                    "client_secret": 'os.environ["IDP_CLIENT_SECRET"]',
                    "scopes": str(env_vars.idp_scopes),
                }
                env["IDP_CLIENT_SECRET"] = env_vars.IDP_CLIENT_SECRET  # type: ignore[assignment]
            else:
                raise ToolkitNotSupported(f"Login flow {env_vars.LOGIN_FLOW} is not supported .")

        handler_args: dict[str, Any] = {}
        if "client" in args:
            handler_args["client"] = "client"
        if "data" in args:
            handler_args["data"] = str(call_args.data)
        if "secrets" in args:
            handler_args["secrets"] = str(function_dict.get("secrets", {}))
        if "function_call_info" in args:
            handler_args["function_call_info"] = str({"local": True})

        load_dotenv = ""
        if (Path.cwd() / ".env").is_file():
            load_dotenv = textwrap.dedent("""
                try:
                    from dotenv import load_dotenv

                    for parent in Path(__file__).resolve().parents:
                        if (parent / ".env").exists():
                            load_dotenv(parent / '.env')
                except ImportError:
                    ...
            """)

        handler_import = self._create_handler_import(handler_file)

        run_check_py = self.run_check_py.format(
            credentials_cls=credentials_cls,
            handler_import=handler_import,
            client_name=CLIENT_NAME,
            project=env_vars.CDF_PROJECT,
            base_url=env_vars.cdf_url,
            credentials_args="\n        ".join(f"{k}={v}," for k, v in credentials_args.items()),
            handler_args="\n        ".join(f"{k}={v}," for k, v in handler_args.items()),
            function_external_id=function_external_id,
            load_dotenv=load_dotenv,
        )
        return run_check_py, env

    @staticmethod
    def _get_function_args(py_file: Path, function_name: str) -> set[str]:
        parsed = ast.parse(safe_read(py_file))
        handle_function = next(
            (item for item in parsed.body if isinstance(item, ast.FunctionDef) and item.name == function_name), None
        )
        if handle_function is None:
            raise ToolkitInvalidFunctionError(f"No {function_name} function found in {py_file}")
        return {a.arg for a in handle_function.args.args}

    @staticmethod
    def _create_handler_import(handler_file: str) -> str:
        return re.sub(r"\.+", ".", handler_file.replace(".py", "").replace("/", ".")).removeprefix(".")


class RunTransformationCommand(ToolkitCommand):
    def run_transformation(self, client: ToolkitClient, external_ids: str | list[str]) -> bool:
        """Run a transformation in CDF"""
        if isinstance(external_ids, str):
            external_ids = [external_ids]
        try:
            transformations: TransformationList = client.transformations.retrieve_multiple(external_ids=external_ids)
        except CogniteAPIError as e:
            print("[bold red]ERROR:[/] Could not retrieve transformations.")
            print(e)
            return False
        if transformations is None or len(transformations) == 0:
            print(f"[bold red]ERROR:[/] Could not find transformation with external_id {external_ids}")
            return False
        for transformation in transformations:
            session = client.iam.sessions.create(session_type="ONESHOT_TOKEN_EXCHANGE")
            if session is None:
                print("[bold red]ERROR:[/] Could not get a oneshot session.")
                return False
            nonce = NonceCredentials(session_id=session.id, nonce=session.nonce, cdf_project_name=client.config.project)
            transformation.source_nonce = nonce
            transformation.destination_nonce = nonce
        try:
            client.transformations.update(transformations)
        except CogniteAPIError as e:
            print("[bold red]ERROR:[/] Could not update transformations with oneshot session.")
            print(e)
            return False
        for transformation in transformations:
            try:
                job = client.transformations.run(transformation_external_id=transformation.external_id, wait=False)
                print(f"Running transformation {transformation.external_id}, status {job.status}...")
            except CogniteAPIError as e:
                print(f"[bold red]ERROR:[/] Could not run transformation {transformation.external_id}.")
                print(e)
        return True


class RunWorkflowCommand(ToolkitCommand):
    def run_workflow(
        self,
        env_vars: EnvironmentVariables,
        organization_dir: Path,
        build_env_name: str | None,
        external_id: str | None,
        version: str | None,
        wait: bool,
    ) -> bool:
        """Run a workflow in CDF"""
        resources = ModuleResources(organization_dir, build_env_name)
        client = env_vars.get_client()
        is_interactive = external_id is None
        workflows = resources.list_resources(WorkflowVersionId, "workflows", WorkflowVersionLoader.kind)
        if len(workflows) == 0:
            raise ToolkitMissingResourceError("No workflows found in modules.")
        if external_id is None:
            # Interactive mode
            choices = [questionary.Choice(title=f"{build.identifier!r}", value=build) for build in workflows]
            selected = questionary.select("Select workflow to run", choices=choices).ask()
        else:
            selected_ = next(
                (
                    build
                    for build in workflows
                    if build.identifier.workflow_external_id == external_id
                    and (version is None or (build.identifier.version == version))
                ),
                None,
            )
            if selected_ is None:
                raise ToolkitMissingResourceError(f"Could not find workflow with external id {external_id}")
            selected = selected_
        id_ = selected.identifier
        triggers = resources.list_resources(str, "workflows", WorkflowTriggerLoader.kind)

        credentials: ClientCredentials | None = None
        input_: dict | None = None
        for trigger in triggers:
            trigger_dict = trigger.load_resource_dict(env_vars.dump(), validate=False)
            if (
                trigger_dict["workflowExternalId"] == id_.workflow_external_id
                and trigger_dict["workflowVersion"] == id_.version
            ):
                print(f"Found trigger {trigger.identifier!r} for workflow {id_!r}")
                if (
                    is_interactive
                    and not questionary.confirm(
                        "Do you want to use input data and authentication from this trigger?"
                    ).ask()
                ):
                    break
                credentials = (
                    ClientCredentials.load(trigger_dict["authentication"]) if "authentication" in trigger_dict else None
                )
                input_ = trigger_dict.get("input")
                break

        try:
            if credentials:
                client = ToolkitClient(
                    config=ToolkitClientConfig(
                        client_name=CLIENT_NAME,
                        project=env_vars.CDF_PROJECT,
                        base_url=env_vars.cdf_url,
                        credentials=OAuthClientCredentials(
                            token_url=env_vars.idp_token_url,
                            client_id=credentials.client_id,
                            client_secret=credentials.client_secret,
                            scopes=env_vars.idp_scopes,
                        ),
                    )
                )
                nonce = client.iam.sessions.create(session_type="ONESHOT_TOKEN_EXCHANGE").nonce
            else:
                nonce = client.iam.sessions.create(session_type="ONESHOT_TOKEN_EXCHANGE").nonce
        except CogniteAPIError as e:
            raise AuthorizationError(f"Could not create oneshot session for workflow {id_!r}: {e!s}") from e

        if is_interactive:
            wait = questionary.confirm("Do you want to wait for the workflow to complete?").ask()
        if id_.version is None:
            raise ToolkitValueError("Version is required for workflow.")
        execution = client.workflows.executions.run(
            workflow_external_id=id_.workflow_external_id,
            version=id_.version,
            input=input_,
            nonce=nonce,
        )
        table = Table(title=f"Workflow {id_!r}")
        table.add_column("Info", justify="left")
        table.add_column("Value", justify="left", style="green")
        table.add_row("Execution id", str(execution.id))
        table.add_row("Status", str(execution.status))
        table.add_row("Created time", f"{ms_to_datetime(execution.start_time or 0):%Y-%m-%d %H:%M:%S}")
        print(table)

        if not wait:
            return True

        workflow = client.workflows.versions.retrieve(id_)
        if workflow is None:
            raise ToolkitMissingResourceError(f"Could not find workflow {id_!r}")

        total = len(workflow.workflow_definition.tasks)
        max_time = sum((task.timeout or 3600) * (task.retries or 3) for task in workflow.workflow_definition.tasks)
        result = cast(WorkflowExecutionDetailed, client.workflows.executions.retrieve_detailed(execution.id))
        with Progress() as progress:
            call_task = progress.add_task("Waiting for workflow execution to complete...", total=total)
            start_time = time.time()
            duration = 0.0
            sleep_time = 1
            while (result is None or result.status.upper() == "RUNNING") and duration < max_time:
                time.sleep(sleep_time)
                sleep_time = min(sleep_time * 2, 15)
                result = cast(
                    WorkflowExecutionDetailed,
                    client.workflows.executions.retrieve_detailed(execution.id),
                )
                duration = time.time() - start_time
                completed_count = sum(
                    1 for task in result.executed_tasks if task.status.upper() not in {"IN_PROGRESS", "SCHEDULED"}
                )
                # Todo remove this print statement. It is added to check why the progress goes early to 100%.
                task_statuses = {task.external_id: task.status for task in result.executed_tasks}
                print(f"Status: {task_statuses}")
                print(f"Complete count: {completed_count}, total: {total}")
                progress.advance(call_task, advance=completed_count)
            progress.advance(call_task, advance=total)
            progress.stop()
        if result is None:
            print(f"Could not find execution {execution.id}")
            return False

        print(f"Workflow {id_!r} execution {execution.id} completed with status {result.status}")

        table = Table(title=f"Workflow Tasks {id_!r}")
        table.add_column("Task")
        table.add_column("Status")
        table.add_column("Start Time")
        table.add_column("End Time")
        table.add_column("Duration")
        table.add_column("ReasonForIncompletion")

        for task in result.executed_tasks:
            task_duration = (
                f"{datetime.timedelta(seconds=(task.end_time - task.start_time) / 1000).total_seconds():.1f} seconds"
                if task.end_time and task.start_time
                else ""
            )
            table.add_row(
                task.external_id,
                task.status,
                f"{ms_to_datetime(task.start_time):%Y-%m-%d %H:%M:%S}" if task.start_time else "",
                f"{ms_to_datetime(task.end_time):%Y-%m-%d %H:%M:%S}" if task.end_time else "",
                task_duration,
                task.reason_for_incompletion,
            )
        print(table)
        return True



from __future__ import annotations

import shutil
from collections import defaultdict
from collections.abc import Callable
from pathlib import Path
from typing import Any

from rich import print
from rich.table import Table

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.exceptions import AuthenticationError, ToolkitValueError
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning
from cognite_toolkit._cdf_tk.utils import read_yaml_file, safe_write
from cognite_toolkit._cdf_tk.utils.file import safe_rmtree, yaml_safe_dump


class ImportTransformationCLI(ToolkitCommand):
    def __init__(
        self,
        get_client: Callable[[], ToolkitClient] | None = None,
        print_warning: bool = True,
        skip_tracking: bool = False,
    ):
        super().__init__(print_warning, skip_tracking)
        self._dataset_external_id_by_id: dict[int, str] = {}
        # We only initialize the client if we need to look up dataset ids.
        self._client: ToolkitClient | None = None
        self._get_client = get_client

    def execute(
        self,
        source: Path,
        destination: Path,
        overwrite: bool,
        flatten: bool,
        clean: bool,
        verbose: bool = False,
    ) -> None:
        # Manifest files are documented at
        # https://cognite-transformations-cli.readthedocs-hosted.com/en/latest/quickstart.html#transformation-manifest
        if source.is_file() and source.suffix in {".yaml", ".yml"}:
            yaml_files = [source]
        elif source.is_file():
            raise ToolkitValueError(f"File {source} is not a YAML file.")
        elif source.is_dir():
            yaml_files = list(source.rglob("*.yaml")) + list(source.rglob("*.yml"))
        else:
            raise ToolkitValueError(f"Source {source} is not a file or directory.")

        if not yaml_files:
            self.warn(LowSeverityWarning("No YAML files found in the source directory."))
            return None

        count_by_resource_type: dict[str, int] = defaultdict(int)
        for yaml_file in yaml_files:
            data = self._load_file(yaml_file)
            if data is None:
                continue

            # The convert schedule and notifications pop off the schedule and notifications
            # keys from the transformation
            schedule = self._convert_schedule(data, data["externalId"], yaml_file)
            notifications = self._convert_notifications(data, data["externalId"], yaml_file)
            transformation, source_query_relative_path = self._convert_transformation(data, yaml_file)
            source_query_path = yaml_file.parent / source_query_relative_path if source_query_relative_path else None

            if source_query_path and not source_query_path.exists():
                raise ToolkitValueError(
                    f"Query file {source_query_path.as_posix()!r} does not exist. "
                    f"This is referenced in {yaml_file.as_posix()!r}."
                )

            if flatten:
                destination_folder = destination
            else:
                destination_folder = destination / yaml_file.relative_to(source).parent
            destination_folder.mkdir(parents=True, exist_ok=True)

            destination_transformation = destination_folder / f"{yaml_file.stem}.Transformation.yaml"
            if not overwrite and destination_transformation.exists():
                self.warn(LowSeverityWarning(f"File already exists at {destination_transformation}. Skipping."))
                continue
            safe_write(destination_transformation, yaml_safe_dump(transformation))
            if source_query_path is not None:
                destination_query_path = destination_folder / f"{destination_transformation.stem}.sql"
                shutil.copyfile(source_query_path, destination_query_path)

            if schedule is not None:
                destination_schedule = destination_folder / f"{yaml_file.stem}.Schedule.yaml"
                safe_write(destination_schedule, yaml_safe_dump(schedule))
            if notifications:
                destination_notification = destination_folder / f"{yaml_file.stem}.Notification.yaml"
                safe_write(destination_notification, yaml_safe_dump(notifications))
            if verbose:
                print(f"Imported {yaml_file} to {destination_folder}.")
            count_by_resource_type["transformation"] += 1
            count_by_resource_type["schedule"] += 1 if schedule is not None else 0
            count_by_resource_type["notification"] += len(notifications)

        print(f"Finished importing from {source} to {destination}.")
        table = Table(title="Import transformation-cli Summary")
        table.add_column("Resource Type", justify="right", style="cyan")
        table.add_column("Count", justify="right", style="magenta")
        for resource_type, count in count_by_resource_type.items():
            table.add_row(resource_type, str(count))
        print(table)

        if clean:
            if source.is_dir():
                print(f"Cleaning up source directory {source}.")
                safe_rmtree(source)
            else:
                print(f"Cleaning up source file {source}.")
                source.unlink()

    def _load_file(self, yaml_file: Path) -> dict[str, Any] | None:
        content = read_yaml_file(yaml_file, expected_output="dict")
        required_keys = {"externalId", "name", "destination", "query"}
        if missing_keys := required_keys - content.keys():
            self.warn(
                LowSeverityWarning(
                    f"Missing required keys {missing_keys} in {yaml_file}. Likely not a Transformation manifest. Skipping."
                )
            )
            return None
        return content

    def _convert_transformation(
        self, transformation: dict[str, Any], source_file: Path
    ) -> tuple[dict[str, Any], Path | None]:
        if "shared" in transformation:
            transformation["isPublic"] = transformation.pop("shared")
        if "action" in transformation:
            transformation["conflictMode"] = transformation.pop("action")
        if "ignoreNullFields" not in transformation:
            # This is required by the API, but the transformation-cli sets it to true by default.
            transformation["ignoreNullFields"] = True
        if isinstance(transformation["destination"], str):
            # The API expects the destination to be a dict with a type key, while the transformation-cli
            # also allows a string.
            transformation["destination"] = {"type": transformation["destination"]}

        source_query_path: Path | None = None
        if isinstance(transformation["query"], dict):
            query = transformation.pop("query")
            if "file" in query:
                source_query_path = Path(query.pop("file"))

        if "dataSetId" in transformation:
            if "dataSetExternalId" in transformation:
                self.warn(LowSeverityWarning(f"Both dataSetId and dataSetExternalId are present in {source_file}."))
            else:
                data_set_external_id = self._lookup_dataset(transformation.pop("dataSetId"))
                if data_set_external_id is None:
                    self.warn(
                        LowSeverityWarning(f"Failed to find DataSet with id {transformation['dataSetId']} in CDF.")
                    )
                else:
                    transformation["dataSetExternalId"] = data_set_external_id

        if "authentication" in transformation:
            authentication = transformation["authentication"]
            if not isinstance(authentication, dict):
                self.warn(LowSeverityWarning(f"Invalid authentication format in {source_file}."))
            else:
                if "tokenUrl" in authentication:
                    authentication["tokenUri"] = authentication.pop("tokenUrl")

                if "read" in authentication or "write" in authentication:
                    # Read or Write in authentication.
                    transformation.pop("authentication")

                read = authentication.pop("read", None)
                if read and isinstance(read, dict):
                    if "tokenUrl" in read:
                        read["tokenUri"] = read.pop("tokenUrl")
                    transformation["sourceOidcCredentials"] = read

                write = authentication.pop("write", None)
                if write and isinstance(write, dict):
                    if "tokenUrl" in write:
                        write["tokenUri"] = write.pop("tokenUrl")
                    transformation["destinationOidcCredentials"] = write
        return transformation, source_query_path

    def _convert_notifications(
        self, transformation: dict[str, Any], external_id: str, source_file: Path
    ) -> list[dict[str, Any]]:
        notifications = []
        if "notifications" in transformation:
            notifications_raw = transformation.pop("notifications")
            if isinstance(notifications_raw, list) and all(isinstance(n, str) for n in notifications_raw):
                notifications = [
                    {"destination": email, "transformationExternalId": external_id} for email in notifications_raw
                ]
            else:
                self.warn(LowSeverityWarning(f"Invalid notifications format in {source_file}."))
        return notifications

    def _convert_schedule(
        self, transformation: dict[str, Any], external_id: str, source_file: Path
    ) -> dict[str, Any] | None:
        schedule: dict[str, Any] | None = None
        if "schedule" in transformation:
            schedule_raw = transformation.pop("schedule")
            if isinstance(schedule_raw, str):
                schedule = {"interval": schedule_raw}
            elif isinstance(schedule_raw, dict):
                schedule = schedule_raw
            else:
                self.warn(LowSeverityWarning(f"Invalid schedule format in {source_file}."))
            if isinstance(schedule, dict):
                schedule["externalId"] = external_id
        return schedule

    def _lookup_dataset(self, dataset_id: int) -> str | None:
        if dataset_id in self._dataset_external_id_by_id:
            return self._dataset_external_id_by_id[dataset_id]
        dataset = self.client.data_sets.retrieve(id=dataset_id)
        if dataset is None or dataset.external_id is None:
            return None
        self._dataset_external_id_by_id[dataset.id] = dataset.external_id
        return dataset.external_id

    @property
    def client(self) -> ToolkitClient:
        if self._client is None:
            if self._get_client is None:
                raise AuthenticationError(
                    "No Cognite Client available. Are you missing a .env file?"
                    "\nThis is required to look up dataset ids in "
                    "the transformation-cli manifest(s)."
                )
            self._client = self._get_client()
        return self._client
